/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* CVC Code Generation for Automatic Generation of Attack Format Strings     */
/*                                                                           */
/* Vinod Ganapathy                                       6 November 2003     */
/*                                                                           */
/* Functionality: Given the set of attack parameters, automatically generate */
/* CVC code for the format string attack.                                    */
/*                                                                           */
/* The code in this file is specifically targeted at printf, and the attack  */
/* that reads from an arbitrary location in memory.                          */
/*                                                                           */
/* The boolean formula whose satisfying inst. gives us a format string that  */
/* can be used for the attack is:                                            */
/*          Ax: format[x] \in {alphabet}       ^                             */
/*          (fmt0 = arg0 + distance)           ^                             */
/*          (fmt0 = 0)                         ^                             */
/*          (arg1 >= arg0 + distance)          ^                             */
/*          (arg1 <= arg0 + distance + fmt_length) ^                         */
/*          arg1 is an integer                 ^                             */
/*          (fmt1 > fmt0 + 1)                  ^                             */
/*          (fmt1 < fmt0 + fmt_length - 1)     ^                             */
/*          fmt1 is an integer                 ^                             */
/*          format[arg1] = adr0                ^                             */
/*          format[arg1+1] = adr0 + 1          ^                             */
/*          format[arg1+2] = adr0 + 2          ^                             */
/*          format[arg1+3] = adr0 + 3          ^                             */
/*          # %s and %x between fmt0 and fmt1 is (arg1 - arg0)/4             */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "printf_read_attack.h"

//-----------------------------------------------------------------------------
// Declarations.
//-----------------------------------------------------------------------------
void printf_read_attack::declarations (void) {
  // The set of format characters for which counters have to be defined.
  std::set<std::string>::iterator fc_set_it;
  
// If you want more uninterpreted functions, insert more alphabets
// into this set. simplified_model.ucl was the code with just 's', full_model
// was the code with 's' and 'x'.
  fc_set.insert("s");
  fc_set.insert("x");

  std::cout << "(*----------------------------------------------------------\n";
  std::cout << " Declarations\n";  
  std::cout << "----------------------------------------------------------*)\n";

  std::cout << "CONST\n\n";

  std::cout << "arg0 : TERM;\n";
  std::cout << "arg1 : TERM;\n";
  std::cout << "fmt0 : TERM;\n";
  std::cout << "fmt1 : TERM;\n";
  std::cout << "distance : TERM;\n";
  std::cout << "fmt_length : TERM;\n";
  std::cout << "adr_lo     : TERM;\n";
  std::cout << "adr_medlo  : TERM;\n";
  std::cout << "adr_medhi  : TERM;\n";
  std::cout << "adr_hi     : TERM;\n";
  std::cout << "format     : FUNC[1];\n";
  std::cout << "percent    : TERM;\n";
  
  for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
    std::cout << *fc_set_it << " : TERM;\n";
    std::cout << *fc_set_it << "_count   :   FUNC[1];\n";  
  }
  
  std::cout << "\n";
  return;
}

//-----------------------------------------------------------------------------
// Initialization code. To set the values of %, s, x, etc. Also initialize the
// attack parameters.
//-----------------------------------------------------------------------------
void printf_read_attack::initialization (void) {
  // The ASCII values for each of the format characters we used!
  // For instance, we were interested in %s, and so we must declare
  std::set<std::string>::iterator fc_set_it;
  int ascii;
  int lc = 1;

  std::cout << "DEFINE\n\n";

//  std::cout << "(*----------------------------------------------------------\n";
//  std::cout << " The code to give concrete values to the attack parameters\n";
//  std::cout << "----------------------------------------------------------*)\n";
  
  if (distance != 0)
  {
    std::cout << "concval" << lc++ << " := distance = succ^" << distance <<"(ZERO);\n\n";
  }
  else
  {
    std::cout << "concval" << lc++ << " := distance = ZERO;\n\n";
  }
  
  if (fmt_length != 0)
  {
    std::cout << "concval" << lc++ << " := fmt_length = succ^" << fmt_length << "(ZERO);\n\n";
  }
  else
  {
    std::cout << "concval" << lc++ << " := fmt_length = ZERO;\n\n";
  }
  
  std::cout << "concval" << lc++ << " := adr_lo = succ^" << address_lsb << "(ZERO);\n\n";
  
  std::cout << "concval" << lc++ << " := adr_medlo = succ^" << address_med_lsb << "(ZERO);\n\n";
  
  std::cout << "concval" << lc++ << " := adr_medhi = succ^" << address_med_msb << "(ZERO);\n\n";
  
  std::cout << "concval" << lc++ << " := adr_hi = succ^" << address_msb << "(ZERO);\n\n";
  
  std::cout << "concval := \n";
  for (int i = 1; i < lc - 1; i++) {
    std::cout << "\tconcval" << i << " &\n";
  }
  
  std::cout << "\tconcval" << lc - 1 << ";\n\n";

//  std::cout << "(*----------------------------------------------------------\n";
//  std::cout << " The code to assign values to the symbols used\n";  
//  std::cout << "----------------------------------------------------------*)\n";
  

  lc = 1;
  // The ASCII for '%' is 37, need to put separately since % is comment in CVC.
  std::cout << "decl" << lc++ << " := percent = succ^37(ZERO);\n";
  
  for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
    ascii = (*fc_set_it)[0];
    std::cout << "decl" << lc++ << " := "<< *fc_set_it << " = succ^" 
              << ascii << "(ZERO);\n";
  }
    
  std::cout << "decl := \n";
  for (int i = 1; i < lc - 1; i++) {
    std::cout << "\tdecl" << i << " &\n";
  }

  std::cout << "\tdecl" << lc - 1 << ";\n";
  std::cout << "\n\n";


  return;
}

//-----------------------------------------------------------------------------
// Attack Encoding. The actual set of boolean constraints that are crucial to
// encode this attack.
//-----------------------------------------------------------------------------
void printf_read_attack::attack_encoding (void) {
  std::set<std::string>::iterator fc_set_it;
  
//  std::cout << "(*----------------------------------------------------------\n";
//  std::cout << " The actual attack\n";
//  std::cout << "----------------------------------------------------------*)\n";
  
  // The following statement is required since we are saying something about
  // format string array only from location 0 onwards in all the other
  // definitions, for instance in the definition of the counters;
  std::cout << "att1 := fmt0 = ZERO;\n\n";
  
  std::cout << "att2 := arg0 = fmt0 - distance;\n\n";
  
  std::cout << "att3 := arg1 >= fmt0;\n\n"; 

  std::cout << "att4 := fmt1 < fmt0 + fmt_length - succ^1(ZERO);\n\n";

  std::cout << "att5 := fmt1 > succ(fmt0);\n\n";

  std::cout << "att6 := arg1 >= arg0 + distance;\n\n";

  std::cout << "att7 := arg1 < arg0 + distance + fmt_length - succ^4(ZERO);\n\n";

  std::cout << "att8 := arg1 = arg0";
  
  for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
    std::cout << " + (4 * " << *fc_set_it << "_count(pred^2(fmt1) - fmt0) )";
  }
  std::cout << ";\n\n";
  
  std::cout << "attack := att1 & att2 & att3 & att4 & att5 & "
            << "att6 & att7 & att8;\n";
}

//-----------------------------------------------------------------------------
// Restrict the contents of the format string to a particular alphabet. 
// In this case, we restrict it to be between values 0-255.
//
// INPUT:  fmt_length
// OUTPUT: format[i] \in alphabet (ASCII in our case) for i \in [0..fmt_length-1]
//-----------------------------------------------------------------------------
void printf_read_attack::format_string_alphabet (void) {
  int i, j;
  std::set<std::string>::iterator fc_set_it;
  int lc = 1;

//  std::cout << "(*----------------------------------------------------------\n";
//  std::cout << " Restricting the alphabet of the format string.\n";
//  std::cout << "----------------------------------------------------------*)\n";
 
/*
// Below is the naive way to restrict the alphabet, saying that the contents
// of each byte in the format string can be anywhere between 0 and 255. This
// leads to disastrous performance.
  for (i = 0; i < fmt_length; i++) {
    std::cout << "ASSERT\n";
    for (j = 0; j < 255; j++) {
      std::cout << "\t\t(format[" << i << "] = " << j << ") OR\n";
    }
    std::cout << "\t\t(format[" << i << "] = 255);\n\n";
  }
*/

// Optimization 1: (Will lead to spurious counterexamples). Make the contents
// anywhere between 0 and 255. That is, remove the integrality constraints, 
// and just add inequalities. This did not lead to performance improvements.

// Optimization 2: Just add constraints saying that the format string just
// contains the different alphabet symbols that are of concern to us. That
// is, it consists only of percent, s, x, etc, and the address that we need
// to read from. This again did not lead to any performance improvements.
  for (i = 0; i < fmt_length; i++) {
  std::cout << "restrict" << lc++ << ":=\n";
    // percent is in the alphabet.
    if (i != 0)
      std::cout << "\t(format(succ^" << i << "(ZERO)) = percent) |\n";
    else
      std::cout << "\t(format(ZERO) = percent) |\n";
      
    // The different formatting characters are in the alphabet.
    for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
      if (i != 0)
      {
        std::cout << "\t(format(succ^" << i << "(ZERO)) = " << *fc_set_it << ") |\n";
      }
      else
      {
        std::cout << "\t(format(ZERO) = " << *fc_set_it << ") |\n";
      }
    }
  
    if (i != 0)
    {
      // The address to be exploited is in the alphabet.
      std::cout << "\t(format(succ^" << i << "(ZERO)) = succ^" 
                                          << address_lsb << "(ZERO)) |\n"
                << "\t(format(succ^" << i << "(ZERO)) = succ^" 
                                          << address_med_lsb << "(ZERO)) |\n"
                << "\t(format(succ^" << i << "(ZERO)) = succ^" 
                                          << address_med_msb << "(ZERO)) |\n"
                << "\t(format(succ^" << i << "(ZERO)) = succ^" 
                                          << address_msb << "(ZERO)) |\n";
       // The don't care value.
      std::cout << "\t(format(succ^" << i << "(ZERO)) = succ^255(ZERO));\n\n";
    }
    else
    {
      // The address to be exploited is in the alphabet.
      std::cout << "\t(format(ZERO) = succ^" 
                                          << address_lsb << "(ZERO)) |\n"
                << "\t(format(ZERO) = succ^" 
                                          << address_med_lsb << "(ZERO)) |\n"
                << "\t(format(ZERO) = succ^" 
                                          << address_med_msb << "(ZERO)) |\n"
                << "\t(format(ZERO) = succ^" 
                                          << address_msb << "(ZERO)) |\n";
       // The don't care value.
      std::cout << "\t(format(ZERO) = succ^255(ZERO));\n\n";

    }
  }

  std::cout << "restrict := \n";
  for (int i = 1; i < lc - 1; i++) {
    std::cout << "\trestrict" << i << " &\n";
  }
  std::cout << "\trestrict" << i << ";\n\n";

  return;
}

//-----------------------------------------------------------------------------
// There are counters for each kind of formatting character, and this function
// is used to generate counter definitions on each of them (for the counter
// that is used in the boolean formula whose satisfyability we want to test)
//
// The counters count the number of formatting characters, for instance %x, %s
// etc. Consider the example of %x: suppose the % is at location 5. x_count(5)
// will include the number of %x's upto and including this %x. x_count(6) will
// trivially be equal to x_count(5) because location 6 has x on it.
// 
// INPUT: fmt_length
// OUTPUT: For each format string character, x, to be counted:
//         Definition of the counter for x_count. x_count(i) counts the number
//         of x's upto and including the i^th location
//-----------------------------------------------------------------------------
void printf_read_attack::counter_definition (void) {
  int i;
  //std::set<std::string> fc_set;
  std::set<std::string>::iterator fc_set_it;
  int lc = 1;

  for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
    std::cout << "(*---------------------------------------------------------\n";
    std::cout << " The counter " << *fc_set_it << "_count is defined below \n";
    std::cout << "---------------------------------------------------------*)\n";

    std::cout << *fc_set_it << "defa0 := ((format(ZERO) = percent) & "
              << "(format(succ(ZERO)) = " << *fc_set_it << "));\n\n";
    std::cout << *fc_set_it << "def0 := \n";
    std::cout << "\t(" << *fc_set_it << "defa0 => (" << *fc_set_it 
              << "_count(ZERO) = succ(ZERO))) &\n";
    std::cout << "\t(~" << *fc_set_it << "defa0  => (" << *fc_set_it 
              << "_count(ZERO) = ZERO));\n\n";

    for (i = 1; i < fmt_length; i++) {
      std::cout << *fc_set_it << "defa" << lc << " := " 
                << "((format(succ^" << i << "(ZERO)) = percent) & "
                << "(format(succ^" << (i + 1) << "(ZERO)) = " 
                << *fc_set_it << "));\n\n";
      
      std::cout << *fc_set_it << "def" << lc << " := \n";
			
			if (i != 1)
			{
      	std::cout << "\t(" << *fc_set_it << "defa" << lc << " => "
        	        << "(" << *fc_set_it << "_count(succ^" << i << "(ZERO)) = "
          	      << *fc_set_it << "_count(succ^" << i - 1 << "(ZERO)) + "
            	    << "succ^1(ZERO))) &\n";
      	std::cout << "\t(~" << *fc_set_it << "defa" << lc << " => "
        	        << "(" << *fc_set_it << "_count(succ^" << i << "(ZERO)) = "
          	      << *fc_set_it << "_count(succ^" << i - 1 << "(ZERO))));\n\n";
      }
			else
			{
     	  std::cout << "\t(" << *fc_set_it << "defa" << lc << " => "
        	        << "(" << *fc_set_it << "_count(succ^" << i << "(ZERO)) = "
          	      << *fc_set_it << "_count(ZERO) + "
            	    << "succ^1(ZERO))) &\n";
      	std::cout << "\t(~" << *fc_set_it << "defa" << lc << " => "
        	        << "(" << *fc_set_it << "_count(succ^" << i << "(ZERO)) = "
          	      << *fc_set_it << "_count(ZERO)));\n\n";

			}
			
			lc++;
    }
    
    std::cout << *fc_set_it << "def := \n";
    for (int i = 0; i < lc - 1 ; i++) {
      std::cout << "\t" << *fc_set_it << "def" << i << " & \n";
    }
    std::cout << "\t" << *fc_set_it << "def" << lc - 1 << ";\n\n";
    lc = 1;
  }

  return;
}

//-----------------------------------------------------------------------------
// Integrality Constraints.
// These are to ensure that the variables fmt1 and arg1 are integers.
// INPUT : fmt_length
// OUTPUT: fmt1 = fmt0 + i for each i \in [0..fmt_length-1]
//         arg1 = arg0 + i for each i \in [0..fmt_length-1-4]
//-----------------------------------------------------------------------------
void printf_read_attack::integrality_constraints (void) {
  int i;

  std::cout << "integral1 :=\n"; 
  for (i = 0; i < fmt_length - 1; i++) {
		if ( i != 0)
    	std::cout << "       (fmt1 = fmt0 + succ^" << i << "(ZERO)) | \n"; 
  	else
    	std::cout << "       (fmt1 = fmt0) | \n"; 
	}
  std::cout << "       (fmt1 = fmt0 + succ^" << fmt_length - 1 << "(ZERO));\n\n";
  
  if (fmt_length - 4 - 1 < 0) {
    std::cerr << "For meaningful results, the fmt_length has to be at least 5\n";
    exit(1);
  }

  std::cout << "integral2 := \n"; 
  for (i = 0; i < fmt_length - 4; i++) {
		if ( i != 0)
    	std::cout << "       (arg1 = fmt0 + succ^" << i << "(ZERO)) |\n"; 
  	else
    	std::cout << "       (arg1 = fmt0) |\n"; 
		
	}
  std::cout << "       (arg1 = fmt0 + succ^" << fmt_length - 4 << "(ZERO));\n\n";
  
  std::cout << "integral := integral1 & integral2;\n";
  return;
}

//-----------------------------------------------------------------------------
// Query string.
// Also encodes particulars of the attack (that the fmt1 char is %s)
//-----------------------------------------------------------------------------
void printf_read_attack::query (void) {
  std::set<std::string>::iterator fc_set_it;

  // The format string grows from smaller addresses to bigger addresses.
  // So format[0] is at a lower address than format[1]
  // Since little endian, have smaller address containing less significant
  // bits
  std::cout << "EXEC\n\n";
  std::cout << "decide( (concval & decl & attack & integral & restrict";
  for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
    std::cout << " & " << *fc_set_it << "def";
  }

  std::cout << ")\n =>\n";
  std::cout << "(~(format(fmt1) = percent) &\n";
  std::cout << "  (format(succ(fmt1)) = s) &\n";
  std::cout << "  (format(arg1) = adr_lo) &\n";
  std::cout << "  (format(succ(arg1)) = adr_medlo) &\n";
  std::cout << "  (format(succ^2(arg1)) = adr_medhi) &\n";
  std::cout << "  (format(succ^3(arg1)) = adr_hi))\n";
  std::cout << "\n);";

  return;
}

//-----------------------------------------------------------------------------
// Generate CVC code for the attack.
//-----------------------------------------------------------------------------
void printf_read_attack::generate_cvc_code (void) {

  std::cout << "MODEL cvctouclid\n";
  std::cout << "CONTROL\n";
  std::cout << "EXTVAR\n";
  std::cout << "STOREVAR\n";
  std::cout << "VAR\n\n\n";

  declarations();
  initialization();
  attack_encoding();
  format_string_alphabet();
  counter_definition();
  integrality_constraints();
  query();

  return;
}
